unit SortTestMain;

(*                                                                             *
  Name:      SortTestMain                                                      *
                                                                               *
  Author:    Ruediger Walter <rue.walter@web.de>                               *
                                                                               *
  Version:   0.4.2 / 2014-11-07                                                *
                                                                               *
             Test application used for the development of TSortableHashList    *
             and the functions CompareShortStringSensitive and                 *
             CompareShortStringInsensitive in unit SortableHashList.           *
             - GUI-parts                                                       *
                                                                               *
             Please see the documentation of the unit TSortableHashList for    *
             basic information.                                                *
                                                                               *
                                                                               *
  Copyright (C) 2014 by Ruediger Walter <rue.walter@web.de>                    *
                                                                               *
  This library is licensed on the same Modifyed LGPL as Free Pascal RTL and    *
  Lazarus LCL. See the file COPYING.FPC, included in your FPC/Lazarus          *
  distribution, for details about the copyright.                               *
                                                                               *
  This program is distributed in the hope that it will be useful, but WITHOUT  *
  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or        *
  FITNESS FOR A PARTICULAR PURPOSE.                                            *
                                                                               *)

{$mode objfpc}{$H-}   // {$H+} would slow down the list loading

interface

uses
  Classes, SysUtils, FileUtil, TAGraph, TASeries, Forms, Controls, Graphics,
  Dialogs, EditBtn, StdCtrls, ExtCtrls, Buttons, Spin, IniFiles,
  SortableHashList, SortTestLog, SortTest_dev, DbgTimer, Math;

type

  TState    = (st_LoadFiles, st_LoadList, st_CalcStat, st_CalcSort, st_SortToStr,
               st_Debugging, st_TimeTest, st_Responsive, st_BigData, st_ListBusy);

  { TMainForm }

  TMainForm = class(TForm)
    Label12 : TLabel;
    PerfTestBtn : TButton;
    CloseBtn : TBitBtn;
    ListBox : TListBox;                // TMemo is much slower on big data
    ShowListBtn : TButton;
    ShowLogBtn : TButton;
    SortToStringsBtn : TButton;
    SysSensBtn : TButton;
    OwnSensBtn : TButton;
    SysInsensBtn : TButton;
    OwnInsensBtn : TButton;
    DebugBtn : TButton;
    FactorInsensEdit : TEdit;
    CycleTimeEdit4 : TEdit;
    Label11 : TLabel;
    Label13 : TLabel;
    Label14 : TLabel;
    Label15 : TLabel;
    NumFilesEdit : TEdit;
    NumLinesEdit : TEdit;
    TotalTimeEdit1 : TEdit;
    CycleTimeEdit1 : TEdit;
    TotalTimeEdit2 : TEdit;
    FactorSensEdit : TEdit;
    CycleTimeEdit2 : TEdit;
    TotalTimeEdit3 : TEdit;
    CycleTimeEdit3 : TEdit;
    TotalTimeEdit4 : TEdit;
    LoadListLabel : TLabel;
    StateLabel : TLabel;
    MaskEdit : TEdit;
    Chart1 : TChart;
    Label10 : TLabel;
    Label6 : TLabel;
    Label7 : TLabel;
    Label8 : TLabel;
    Label9 : TLabel;
    MatchSeries : TLineSeries;
    MismatchSeries : TLineSeries;
    Panel2 : TPanel;
    Panel3 : TPanel;
    Label2 : TLabel;
    Label3 : TLabel;
    Label4 : TLabel;
    Label5 : TLabel;
    NumFilesLabel : TLabel;
    Panel1 : TPanel;
    PathEdit : TDirectoryEdit;
    Label1 : TLabel;
    RunSpin : TSpinEdit;
    procedure SortToStringsBtnClick(Sender : TObject);
    procedure ShowLogBtnClick(Sender : TObject);
    procedure SortBtnClick(Sender : TObject);
    procedure DebugBtnClick(Sender : TObject);
    procedure FormActivate(Sender : TObject);
    procedure ListBoxSelectionChange(Sender : TObject; User : boolean);
    procedure MaskEditEditingDone(Sender : TObject);
    procedure RunSpinChange(Sender : TObject);
    procedure ShowListBtnClick(Sender : TObject);
    procedure FormCreate(Sender : TObject);
    procedure FormDestroy(Sender : TObject);
    procedure PathEditChange(Sender : TObject);
    procedure PerfTestBtnClick(Sender : TObject);
  private
    IniFileName,
    LogFileName : String;
    IniFile     : TIniFile;
    LogList     : TStringList;
    HashList    : TTestSortHashList;
    Mask        : String;
    SortTime    : Array[ft_ShortSens..ft_SysInsens] of Extended;
    n_cp        : Int64;
    ChartDisplayed : TStatKind;
    n_st,
    NumFiles    : LongInt;
    ListLoading,
    IsOnScreen  : Boolean;
    FState      : TState;
    procedure RunSort(FT : TFuncType);
    procedure SetState(NewState : TState);
    procedure ResetTimes;
    procedure NewLogItem(DoWhat : String);
    procedure ShowChart(SK : TStatKind);
    property  State: TState read FState write SetState;
  end;

var
  MainForm : TMainForm;

implementation

{$R *.lfm}

const
  StartPath   = '/usr/share/lazarus/1.2.4/';
  StartMask   = '*.pas;*.pp;*.p;*.inc;*.lpr';
  SessionSect = 'LastSession';
  PathIdent   = 'Path';
  MaskIdent   = 'Mask';
  RunsIdent   = 'Runs';
  LogFileExt  = '.log';
  IniFileExt  = '.cfg';

{ TMainForm }

{ **************************** Create / Destroy ****************************** }

procedure TMainForm.FormCreate(Sender : TObject);
var
  s : String;
begin
  HashList := TTestSortHashList.Create(CaseSensitive);
  s := Application.ExeName;
  IniFileName := ChangeFileExt(s, IniFileExt);
  LogFileName := ChangeFileExt(s, LogFileExt);
  IniFile := TIniFile.Create(IniFileName);
  LogList := TStringList.Create;
  if FileExists(LogFileName) then LogList.LoadFromFile(LogFileName);
  if DirectoryExists(StartPath) then s := StartPath else s := '';
  PathEdit.Directory := IniFile.ReadString (SessionSect, PathIdent, s);
  Mask := IniFile.ReadString (SessionSect, MaskIdent, StartMask);
  RunSpin.Value := IniFile.ReadInteger(SessionSect, RunsIdent, 1);
  ResetTimes;
  IsOnScreen := false;
  ListLoading := false;
  MaskEdit.Text := Mask;
  ChartDisplayed := sk_Tmp;           s := copy(s,1,1);
end;

procedure TMainForm.FormDestroy(Sender : TObject);
begin
  IniFile.WriteString (SessionSect, PathIdent, PathEdit.Directory);
  IniFile.WriteString (SessionSect, MaskIdent, MaskEdit.Text);
  IniFile.WriteInteger(SessionSect, RunsIdent, RunSpin.Value);
  LogList.SaveToFile(LogFileName);
  LogList.Free;
  IniFile.Free;
  HashList.Free;
end;

{ ******************************* Internals ********************************** }

procedure TMainForm.FormActivate(Sender : TObject);
begin
  Application.ProcessMessages;
  SetState(st_LoadFiles);
  IsOnScreen := true;
  PathEditChange(self);
end;

procedure TMainForm.ResetTimes;
begin
  FillChar(SortTime,SizeOf(SortTime),0);
  TotalTimeEdit1.Clear;
  TotalTimeEdit2.Clear;
  TotalTimeEdit3.Clear;
  TotalTimeEdit4.Clear;
  CycleTimeEdit1.Clear;
  CycleTimeEdit2.Clear;
  CycleTimeEdit3.Clear;
  CycleTimeEdit4.Clear;
  FactorSensEdit.Clear;
  FactorInsensEdit.Clear;
end;

procedure TMainForm.SetState(NewState : TState);
begin
  FState := NewState;
  if NewState = st_Responsive then
  begin
    LoadListLabel.Caption := '';
    StateLabel.Font.Color := clGreen;
    SysSensBtn.Enabled    := true;
    OwnSensBtn.Enabled    := true;
    SysInsensBtn.Enabled  := true;
    OwnInsensBtn.Enabled  := true;
    ShowListBtn.Enabled   := true;
    DebugBtn.Enabled      := true;
    PerfTestBtn.Enabled   := true;
    ShowLogBtn.Enabled    := true;
    SortToStringsBtn.Enabled   := true;
  end
  else begin
    StateLabel.Font.Color := clRed;
    SysSensBtn.Enabled    := false;
    OwnSensBtn.Enabled    := false;
    SysInsensBtn.Enabled  := false;
    OwnInsensBtn.Enabled  := false;
    ShowListBtn.Enabled   := false;
    DebugBtn.Enabled      := false;
    PerfTestBtn.Enabled   := false;
    ShowLogBtn.Enabled    := false;
    SortToStringsBtn.Enabled   := false;
  end;
  case NewState of
    st_LoadFiles  : StateLabel.Caption := 'Loading files.......';
    st_LoadList   : StateLabel.Caption := 'Loading ListBox..............';
    st_CalcStat   : StateLabel.Caption := 'Grabbing Statistics.......';
    st_CalcSort   : StateLabel.Caption := 'Sorting.......';
    St_SortToStr  : StateLabel.Caption := 'Sort to Strings.......';
    st_Debugging  : StateLabel.Caption := 'Debugging.......';
    st_TimeTest   : StateLabel.Caption := 'Measure Timings..............';
    st_Responsive : StateLabel.Caption := 'Responsive';
    st_BigData    : LoadListLabel.Caption := 'Get a cup of coffee, this sucks!';
    st_ListBusy   : LoadListLabel.Caption := LoadListLabel.Caption  + #10 + #10 +
                                          'Still loading..............';
  end;
  Application.ProcessMessages;
end;

{ ******************************* User input ********************************* }

procedure TMainForm.RunSpinChange(Sender : TObject);
begin
  ResetTimes;
end;

procedure TMainForm.MaskEditEditingDone(Sender : TObject);
begin
  if MaskEdit.Text = Mask then exit;
  Mask := MaskEdit.Text;
  PathEditChange(self);
end;

procedure TMainForm.PathEditChange(Sender : TObject);
var
  FileList : TStringList;

  {Sub-}
  procedure LoadFilesToHashList;
  var
    i, k, n1, n2, dup : LongInt;
    s    : ShortString;
    SL   : TStringList;
  begin
    SL := TStringList.Create;
    HashList.Clear;
    dup := 0;
    for i := 0 to FileList.Count -1 do
    begin
      SL.Clear;
      SL.LoadFromFile(FileList[i]);
      for k := 0 to SL.Count -1 do
      begin
        s := SL[k];                           // Conversion to ShortString
        if s = '' then continue;
        n1 := 1;                              // delete tabs and spaces from both sides
        n2 := Length(s);
        while (s[n1] <= ' ') and (n1 <= n2) do inc(n1);
        while (s[n2] <= ' ') and (n1 <= n2) do dec(n2);
        if n1 > n2 then continue;
        if (n1 > 1) or (n2 < Length(s)) then s := copy(s, n1, n2-n1+1);
        if HashList.FindIndexOf(s) < 0 then HashList.Add(s, self)
        else inc(dup); // lots of duplicates, but dup=0, is this a bug in HashList.FindIndexOf() ?
      end;             // Solved & Bug found: Data pointer must not be nil.
    end;               // http://bugs.freepascal.org/view.php?id=26800
    SL.Free;
  end;
  {/Sub}

begin
  if not (IsOnScreen and Assigned(HashList) and DirectoryExists(PathEdit.Directory)) then exit;
  SetState(st_LoadFiles);
  ResetTimes;
  FileList := FindAllFiles(PathEdit.Directory, MaskEdit.Text, true); // FileList automatically created
  NumFiles := FileList.Count;
  NumFilesEdit.Text := IntToStr(NumFiles);
  LoadFilesToHashList;
  n_st := HashList.Count;
  NumLinesEdit.Text := IntToStr(HashList.Count);
  SetState(st_CalcStat);
  HashList.SortStatistics;
  SetState(st_Responsive);
  if Assigned(FileList) then FileList.Free;
  ShowChart(sk_Tmp);
  NewLogItem('New Data:');
  LogList.Add('Directory:       ' + PathEdit.Directory);
  LogList.Add('Filename Mask:   ' + MaskEdit.Text);
  LogList.Add('Number of Files: ' + NumFilesEdit.Text);
  LogList.Add('Number of Lines: ' + NumLinesEdit.Text);
end;

{ ******************************* Sort Test ********************************** }

procedure TMainForm.RunSort(FT : TFuncType);
var
  i : LongInt;
  s : String;
begin
  SetState(st_CalcSort);
  if FT in [ft_ShortSens, ft_SysSens] then
    n_cp := HashList.Statistics[sk_SortSens].NumCompares * RunSpin.Value
  else n_cp := HashList.Statistics[sk_SortInsens].NumCompares * RunSpin.Value;
  case FT of
    ft_ShortSens   : HashList.CaseSensitivity := CaseSensitive;
    ft_ShortInsens : HashList.CaseSensitivity := CaseInsensitive;
  end;
  DebugTimerStart;
  For i := 1 to RunSpin.Value do
    case FT of
      ft_ShortSens   : HashList.Sort;
      ft_ShortInsens : HashList.Sort;
      ft_SysSens     : HashList.SortSysCompare(CaseSensitive);
      ft_SysInsens   : HashList.SortSysCompare(CaseInsensitive);
    end;
  SortTime[FT] := DebugTimerStop(n_cp);
  case FT of
    ft_ShortSens   : begin
                       s := 'SysUtils.CompareStr';
                       TotalTimeEdit2.Text := DebugTimeTotal;
                       CycleTimeEdit2.Text := DebugTimePerCycle;
                     end;
    ft_ShortInsens : begin
                       s := 'CompareShortStringSensitive';
                       TotalTimeEdit4.Text := DebugTimeTotal;
                       CycleTimeEdit4.Text := DebugTimePerCycle;
                     end;
    ft_SysSens     : begin
                       s := 'SysUtils.CompareText';
                       TotalTimeEdit1.Text := DebugTimeTotal;
                       CycleTimeEdit1.Text := DebugTimePerCycle;
                     end;
    ft_SysInsens   : begin
                       s := 'CompareShortStringInsensitive';
                       TotalTimeEdit3.Text := DebugTimeTotal;
                       CycleTimeEdit3.Text := DebugTimePerCycle;
                     end;
  end;
  if HashList.CaseSensitivity = CaseSensitive then ShowChart(sk_SortSens)
  else ShowChart(sk_SortInsens);
  NewLogItem('Sort Test: ' + s);
  LogList.Add('Number of Strings  ' + IntToStr(n_St));
  LogList.Add('Number of Runs:    ' + IntToStr(RunSpin.Value));
  LogList.Add('Comparisons / Run: ' + IntToStr(n_Cp div RunSpin.Value));
  LogList.Add('Total Time:        ' + DebugTimeTotal);
  LogList.Add('Time / Cycle:      ' + DebugTimePerCycle);
  if (SortTime[ft_ShortSens] > 0) and (SortTime[ft_SysSens] > 0) then
    FactorSensEdit.Text := format('%.2f',[SortTime[ft_SysSens] / SortTime[ft_ShortSens]])
  else FactorSensEdit.Text := '';
  if (SortTime[ft_ShortInsens] > 0) and (SortTime[ft_SysInsens] > 0) then
    FactorInsensEdit.Text := format('%.2f',[SortTime[ft_SysInsens] / SortTime[ft_ShortInsens]])
  else FactorInsensEdit.Text := '';
  if HashList.CaseSensitivity = CaseSensitive then s := FactorSensEdit.Text
  else s := FactorInsensEdit.Text;
  if s <> '' then LogList.Add('Factor:            ' + s);
  SetState(st_Responsive);
end;

procedure TMainForm.SortBtnClick(Sender : TObject);
begin
  if Sender = SysSensBtn then RunSort(ft_SysSens)
  else if Sender = OwnSensBtn then RunSort(ft_ShortSens)
    else if Sender = SysInsensBtn then RunSort(ft_SysInsens)
      else if Sender = OwnInsensBtn then RunSort(ft_ShortInsens);
end;

{ ********************** Debugging / Performance Test ************************ }

procedure TMainForm.DebugBtnClick(Sender : TObject);
const
  Txt1  = '.......Done!     Comparisons: ';
  Txt2  = '   /   Errors: ';
begin
  SetState(st_Debugging);
  ListBox.Clear;
  ListBox.AddItem('Running a debug version of TSortableHashList with both sensitive ' +
                  'and insensitive compare functions. ', nil);
  ListBox.AddItem('The result of each comparison by a debugged function is compared ' +
                  'to the result of its SysUtils ', nil);
  ListBox.AddItem('counterpart (with the same comparison strings), and a 10 times ' +
                  'loop of that comparison will be', nil);
  ListBox.AddItem('triggered if a difference is found. To investigate errors (if ' +
                  'there are any) a Breakpoint should', nil);
  ListBox.AddItem('be set in SortTest_dev.TTestSortHashList.DebugCompare.' + #10, nil);
  ListBox.AddItem('Calculating Case-Sensitive Compare.........', nil);
  Application.ProcessMessages;
  HashList.SortDebug(CaseSensitive);
  ListBox.AddItem(Txt1 + IntToStr(HashList.Statistics[sk_Tmp].NumCompares) +
                  Txt2 + IntToStr(HashList.Statistics[sk_Tmp].NumErrors) + #10, nil);
  ListBox.AddItem('Calculating Case-Insensitive Compare.........', nil);
  Application.ProcessMessages;
  HashList.SortDebug(CaseInsensitive);
  ListBox.AddItem(Txt1 + IntToStr(HashList.Statistics[sk_Tmp].NumCompares) +
                  Txt2 + IntToStr(HashList.Statistics[sk_Tmp].NumErrors) + #10, nil);
  NewLogItem('Debug:');
  LogList.AddStrings(ListBox.Items);
  SetState(st_Responsive);
end;

procedure TMainForm.PerfTestBtnClick(Sender : TObject);
const
  TenE6 = 1000000;
  TenE9 = 1000000000;
var
  SysTimePerLoop,
  TotalWrapTime,
  RealCompTime,
  WrapTimePerLoop,
  CompTimePerLoop : Extended;
  CodeVersion     : String;

  {Sub-}
  procedure ScrollDown(LF: Boolean);
  begin
    if LF then ListBox.AddItem('', nil);
    Application.ProcessMessages;
    ListBox.ItemIndex := ListBox.Count-1;
    Application.ProcessMessages;
  end;
  {/Sub}

  {Sub-}
  procedure WriteTimings;
  begin
    TotalWrapTime   := WrapTimePerLoop * HashList.Statistics[sk_Tmp].NumCompares;
    RealCompTime    := HashList.Statistics[sk_Tmp].TotalTime - TotalWrapTime;
    CompTimePerLoop := RealCompTime * TenE9 / HashList.Statistics[sk_Tmp].NumCompares;
    ListBox.AddItem('......Done!' + #10 +
      '      Comparisons:              ' +
        IntToStr(HashList.Statistics[sk_Tmp].NumCompares div TenE6) + ' Millions' + #10 +
      '      Total Time:               ' +
        format('%.3f',[HashList.Statistics[sk_Tmp].TotalTime]) + ' s' + #10 +
      '      Real Compare Time:        ' + format('%.3f',[RealCompTime]) + ' s' + #10 +
      '      Real Compare Time / Loop: ' + format('%.1f',[CompTimePerLoop]) + ' ns', nil);
    ScrollDown(true);
  end;
  {/Sub}

begin
  SetState(st_TimeTest);
  SetState(st_BigData);
  CodeVersion :=                    '       Pascal-Version        ';
  {$ifdef CPUX86_64} CodeVersion := ' CPUX86_64 Assembler-Version '; {$endif}
  {$ifdef CPUI386}   CodeVersion := '  CPUI386 Assembler-Version  '; {$endif}
  ListBox.Clear;
  ListBox.AddItem('Testing the performance of both sensitive and insensitive ' +
                  'compare functions and their SysUtils conterparts.', nil);
  ListBox.AddItem('This test will take about 7 minutes.', nil);
  ListBox.AddItem('First the "Wrap-Time" will be computed. This means the time ' +
                  'part the test routine needs to prepare', nil);
  ListBox.AddItem('the Compare function call including the call itself (with ' +
                  'the preliminaries of "begin" and "end"', nil);
  ListBox.AddItem('inside the compare function). What is called the "Real ' +
                  'Compare Time" is the time part inside the', nil);
  ListBox.AddItem('Compare function between "Begin" and "end" - that is what ' +
                  'the function really does inside.', nil);
  ListBox.AddItem('To isolate this time part outside the Compare function ' +
                  'an empty function (just a "begin end")', nil);
  ListBox.AddItem('but with the same parameters und function result will be ' +
                  'called by the testing routine.', nil);
  ListBox.AddItem('Each run begins with a rough time test and then a loop of ' +
                  'about one minute will be started.', nil);
  ListBox.AddItem('There is no micro time measurement - only the Total Time ' +
                  'is measured, all other values are calculated.', nil);
  ListBox.AddItem('The results of this test are not directly comparable to those ' +
                  'of the sort test. Here about 95 % of all', nil);
  ListBox.AddItem('comparisons will mismatch on the first char, while the portion ' +
                  'of first-char decisions is only about 25 %', nil);
  ListBox.AddItem('on sorting (due to the sorting algorithm). Also the sort test ' +
                  'does not exlude the "wrap-time".', nil);
  ListBox.AddItem('', nil);
  { ********* Statistics Case-Sensitive ********* }
  ListBox.AddItem('Grabbing distribution statistics case-sensitive...........', nil);
  ScrollDown(false);
  HashList.TestPerformance(ft_StatSens, 20);
  ListBox.AddItem('......Done!', nil);
  ListBox.AddItem('', nil);
  ShowChart(sk_EquiDistSens);
  { ********* Wrap time ********* }
  ListBox.AddItem('Running wraptime-part evaluation (with dummy Compare ' +
                  'function)  ...........', nil);
  ScrollDown(false);
  HashList.TestPerformance(ft_Empty, 20);
  WrapTimePerLoop := HashList.Statistics[sk_Tmp].TotalTime / HashList.Statistics[sk_Tmp].NumCompares;
  ListBox.AddItem('......Done!' + #10 +
    '      Comparisons:              ' + IntToStr(HashList.Statistics[sk_Tmp].NumCompares div TenE6) +
    ' Millions' + #10 +
    '      Total Time:               ' + format('%.3f',[HashList.Statistics[sk_Tmp].TotalTime]) +
    ' s' + #10 +
    '      Wrapping Time / Loop:     ' + format('%.1f',[WrapTimePerLoop * TenE9]) + ' ns', nil);
  ListBox.AddItem('', nil);
  { ********* Sys Compare Case-Sensitive ********* }
  ListBox.AddItem('            Running Case-Sensitive Test............', nil);
  ListBox.AddItem('', nil);
  ListBox.AddItem('**********      SysUtils.CompareStr      **********', nil);
  ScrollDown(false);
  HashList.TestPerformance(ft_SysSens, 60);
  WriteTimings;
  SysTimePerLoop := CompTimePerLoop;
  { ********* Short Compare Case-Sensitive ********* }
  ListBox.AddItem('**********  CompareShortStringSensitive  **********', nil);
  ListBox.AddItem('********** ' + CodeVersion + ' **********', nil);
  ScrollDown(false);
  HashList.TestPerformance(ft_ShortSens, 60);
  WriteTimings;
  ListBox.AddItem('      Performance Factor:       ' +
                  format('%.2f',[SysTimePerLoop / CompTimePerLoop]), nil);
  ListBox.AddItem('', nil);
  { ********* Statistics Case-Insensitive ********* }
  ListBox.AddItem('Grabbing distribution statistics case-insensitive...........', nil);
  ScrollDown(false);
  HashList.TestPerformance(ft_StatInsens, 20);
  ListBox.AddItem('......Done!', nil);
  ListBox.AddItem('', nil);
  ShowChart(sk_EquiDistInsens);
  { ********* Sys Compare Case-Insensitive ********* }
  ListBox.AddItem('           Running Case-Insensitive Test...........', nil);
  ListBox.AddItem('', nil);
  ListBox.AddItem('**********      SysUtils.CompareText     **********', nil);
  ScrollDown(false);
  HashList.TestPerformance(ft_SysInsens, 60);
  WriteTimings;
  SysTimePerLoop := CompTimePerLoop;
  { ********* Short Compare Case-Insensitive ********* }
  ListBox.AddItem('********** CompareShortStringInsensitive **********', nil);
  ListBox.AddItem('********** ' + CodeVersion + ' **********', nil);
  ScrollDown(false);
  HashList.TestPerformance(ft_ShortInsens, 60);
  WriteTimings;
  ListBox.AddItem('      Performance Factor:       ' +
                  format('%.2f',[SysTimePerLoop / CompTimePerLoop]), nil);
  NewLogItem('Performance Test:');
  LogList.AddStrings(ListBox.Items);
  SetState(st_Responsive);
end;

{ ***************************** Chart / ListBox ****************************** }

procedure TMainForm.ShowChart(SK : TStatKind);
var
  i : LongInt;
  d : double;
  sum11_15, sum16_20, sum21_30,
  sum31_50, sum51_100, sum101_255,
  goodness : double;
begin
  if ChartDisplayed = SK then exit;
  ChartDisplayed := SK;
  MatchSeries.Clear;
  MismatchSeries.Clear;
  if SK = sk_Tmp then
  begin
    Label6.Caption  := '';
    Label7.Caption  := '';
    Label8.Caption  := '';
    Label9.Caption  := '';
    Label10.Caption := '';
    exit;
  end;
  n_cp := HashList.Statistics[SK].NumMismatches + HashList.Statistics[SK].NumMatches;
  for i := 1 to 100 do
  begin
    d := HashList.Statistics[SK].Matches[i]*100/HashList.Statistics[SK].NumMatches;
    Matchseries.AddXY(i,d);
    d := HashList.Statistics[SK].Mismatches[i]*100/HashList.Statistics[SK].NumMismatches;
    Mismatchseries.AddXY(i,d);
  end;
  goodness := n_st * log2(n_st) / n_cp;
  ThousandSeparator := ' ';
  if (SK = sk_SortSens) or (SK = sk_EquiDistSens) then
       Label8.Caption := 'Case-Sensitive Compare:'
  else Label8.Caption := 'Case-Insensitive Compare:';
  Label9.Caption := 'Number of Strings:     N_st = ' + format('%11.0n',[n_st/1]) + #10 +
                    'Number of Comparisons: N_cp = ' + format('%11.0n',[N_cp/1]) + #10 +
                    'Goodness:  N_st * log2(N_st) / N_cp = ' + format('%.2f',[Goodness]);
  with HashList.Statistics[SK] do
  begin
    sum11_15   := 0; for i :=  11 to  15 do sum11_15   := sum11_15   + Mismatches[i];
    sum16_20   := 0; for i :=  16 to  20 do sum16_20   := sum16_20   + Mismatches[i];
    sum21_30   := 0; for i :=  21 to  30 do sum21_30   := sum21_30   + Mismatches[i];
    sum31_50   := 0; for i :=  31 to  50 do sum31_50   := sum31_50   + Mismatches[i];
    sum51_100  := 0; for i :=  51 to 100 do sum51_100  := sum51_100  + Mismatches[i];
    sum101_255 := 0; for i := 101 to 255 do sum101_255 := sum101_255 + Mismatches[i];
    Label6.Caption := 'Mismatches:' + #10 +
          IntToStr(NumMismatches) + ' = ' +
          format('%.2f',[NumMismatches * 100 / (NumMismatches + NumMatches)]) + ' %' + #10 + #10 +
          '    Pos       %' + #10 +
          '        1   ' + format('%5.2f',[Mismatches[1] * 100 / NumMismatches]) + #10 +
          '        2   ' + format('%5.2f',[Mismatches[2] * 100 / NumMismatches]) + #10 +
          '        3   ' + format('%5.2f',[Mismatches[3] * 100 / NumMismatches]) + #10 +
          '        4   ' + format('%5.2f',[Mismatches[4] * 100 / NumMismatches]) + #10 +
          '        5   ' + format('%5.2f',[Mismatches[5] * 100 / NumMismatches]) + #10 +
          '        6   ' + format('%5.2f',[Mismatches[6] * 100 / NumMismatches]) + #10 +
          '        7   ' + format('%5.2f',[Mismatches[7] * 100 / NumMismatches]) + #10 +
          '        8   ' + format('%5.2f',[Mismatches[8] * 100 / NumMismatches]) + #10 +
          '        9   ' + format('%5.2f',[Mismatches[9] * 100 / NumMismatches]) + #10 +
          '       10   ' + format('%5.2f',[Mismatches[10]* 100 / NumMismatches]) + #10+
          ' 11 -  15   ' + format('%5.2f',[sum11_15      * 100 / NumMismatches]) + #10+
          ' 16 -  20   ' + format('%5.2f',[sum16_20      * 100 / NumMismatches]) + #10 +
          ' 21 -  30   ' + format('%5.2f',[sum21_30      * 100 / NumMismatches]) + #10 +
          ' 31 -  50   ' + format('%5.2f',[sum31_50      * 100 / NumMismatches]) + #10 +
          ' 51 - 100   ' + format('%5.2f',[sum51_100     * 100 / NumMismatches]) + #10 +
          '101 - 255   ' + format('%5.2f',[sum101_255    * 100 / NumMismatches]);
    sum11_15   := 0; for i :=  11 to  15 do sum11_15   := sum11_15   + Matches[i];
    sum16_20   := 0; for i :=  16 to  20 do sum16_20   := sum16_20   + Matches[i];
    sum21_30   := 0; for i :=  21 to  30 do sum21_30   := sum21_30   + Matches[i];
    sum31_50   := 0; for i :=  31 to  50 do sum31_50   := sum31_50   + Matches[i];
    sum51_100  := 0; for i :=  51 to 100 do sum51_100  := sum51_100  + Matches[i];
    sum101_255 := 0; for i := 101 to 255 do sum101_255 := sum101_255 + Matches[i];
    Label7.Caption := 'Matches:' + #10 +
          IntToStr(NumMatches) + ' = ' +
          format('%.2f',[NumMatches * 100 / (NumMismatches + NumMatches)]) + ' %' + #10 + #10 +
          '  Length      %' + #10 +
          '        1   ' + format('%5.2f',[Matches[1] * 100 / NumMatches]) + #10 +
          '        2   ' + format('%5.2f',[Matches[2] * 100 / NumMatches]) + #10 +
          '        3   ' + format('%5.2f',[Matches[3] * 100 / NumMatches]) + #10 +
          '        4   ' + format('%5.2f',[Matches[4] * 100 / NumMatches]) + #10 +
          '        5   ' + format('%5.2f',[Matches[5] * 100 / NumMatches]) + #10 +
          '        6   ' + format('%5.2f',[Matches[6] * 100 / NumMatches]) + #10 +
          '        7   ' + format('%5.2f',[Matches[7] * 100 / NumMatches]) + #10 +
          '        8   ' + format('%5.2f',[Matches[8] * 100 / NumMatches]) + #10 +
          '        9   ' + format('%5.2f',[Matches[9] * 100 / NumMatches]) + #10 +
          '       10   ' + format('%5.2f',[Matches[10]* 100 / NumMatches]) + #10+
          ' 11 -  15   ' + format('%5.2f',[sum11_15   * 100 / NumMatches]) + #10+
          ' 16 -  20   ' + format('%5.2f',[sum16_20   * 100 / NumMatches]) + #10 +
          ' 21 -  30   ' + format('%5.2f',[sum21_30   * 100 / NumMatches]) + #10 +
          ' 31 -  50   ' + format('%5.2f',[sum31_50   * 100 / NumMatches]) + #10 +
          ' 51 - 100   ' + format('%5.2f',[sum51_100  * 100 / NumMatches]) + #10 +
          '101 - 255   ' + format('%5.2f',[sum101_255 * 100 / NumMatches]);
    Label10.Caption := 'Test data:' + #10 + #10 +
          'All not-empty && not-duplicated lines of' + #10 +
          'all  ' + MaskEdit.Text + '  Files in' + #10 +
          PathEdit.Directory + #10 + #10 +
          '-  ' + intToStr(NumFiles) + ' Files' + #10 +
          '-  ' + intToStr(HashList.Count) + ' Strings' + #10 +
          '-  Leading and trailing tabs and spaces' + #10 +
          '   removed';
  end;
  NewLogItem('Data Distribution:');
  LogList.Add(Label8.Caption);
  LogList.Add(Label9.Caption);
  LogList.Add('');
  LogList.Add(Label6.Caption);
  LogList.Add('');
  LogList.Add(Label7.Caption);
  Application.ProcessMessages;
end;

procedure TMainForm.ListBoxSelectionChange(Sender : TObject; User : boolean);
begin                      // recursive call until the end of ListBox loading
  if not ListLoading or (State < st_BigData) then exit;
  if ListBox.Count < HashList.Count then ListBox.ItemIndex := ListBox.Count-1
  else begin
    SetState(st_Responsive);
    ListLoading := false;
  end;
end;

procedure TMainForm.ShowListBtnClick(Sender : TObject);
var
  i : LongInt;
begin
  SetState(st_LoadList);
  ListLoading := true;
  if HashList.Count > 200000 then SetState(st_BigData);
  ListBox.Clear;
  Application.ProcessMessages;
  if Length(HashList.SortArray) > 0 then
    for i := 0 to HashList.Count -1 do ListBox.Items.Add(HashList.SortArray[i]^)
  else for i := 0 to HashList.Count -1 do ListBox.Items.Add(HashList.NameOfIndex(i));
  if State = st_BigData then
  begin
    SetState(st_ListBusy);
    ListBox.ItemIndex := ListBox.Count-1;
  end;
  ListBox.ScrollWidth := 2000;
  if State = st_LoadList then SetState(st_Responsive);
end;

procedure TMainForm.SortToStringsBtnClick(Sender : TObject);
begin
  SetState(st_SortToStr);
  ListLoading := true;
  if HashList.Count > 200000 then SetState(st_BigData);
  ListBox.Clear;
  Application.ProcessMessages;
  HashList.SortToStrings(ListBox.Items, NoObjects);
  if State = st_BigData then
  begin
    SetState(st_ListBusy);
    ListBox.ItemIndex := ListBox.Count-1;
  end;
  ListBox.ScrollWidth := 2000;
  if State < st_Responsive then SetState(st_Responsive);
end;

{ ********************************* LogFile ********************************** }

procedure TMainForm.NewLogItem(DoWhat : String);
var
  s : String;
begin
  s := '------------------------------';
  s := s + '  ' + DoWhat + '  ' + DateTimeToStr(Now) + '  ' + s;
  LogList.Add('');
  LogList.Add(s);
  LogList.Add('');
end;

procedure TMainForm.ShowLogBtnClick(Sender : TObject);
begin
  LogForm.LogMemo.Lines.Assign(LogList);
  LogForm.ShowModal;
  LogList.Assign(LogForm.LogMemo.Lines);        // if cleared in LogForm
end;


end.

