unit SortTest_dev;

(*                                                                             *
  Name:      SortTest_dev                                                      *
                                                                               *
  Author:    Ruediger Walter <rue.walter@web.de>                               *
                                                                               *
  Version:   0.4.2 / 2014-11-07                                                *
                                                                               *
             Test application used for the development of TSortableHashList    *
             and the functions CompareShortStringSensitive and                 *
             CompareShortStringInsensitive in unit SortableHashList.           *
             - Non-GUI-parts                                                   *
                                                                               *
             Please see the documentation of the unit TSortableHashList for    *
             basic information.                                                *
                                                                               *
                                                                               *
  Copyright (C) 2014 by Ruediger Walter <rue.walter@web.de>                    *
                                                                               *
  This library is licensed on the same Modifyed LGPL as Free Pascal RTL and    *
  Lazarus LCL. See the file COPYING.FPC, included in your FPC/Lazarus          *
  distribution, for details about the copyright.                               *
                                                                               *
  This program is distributed in the hope that it will be useful, but WITHOUT  *
  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or        *
  FITNESS FOR A PARTICULAR PURPOSE.                                            *
                                                                               *)


{$mode objfpc}

interface

uses
  Classes, SysUtils, SortableHashList, DbgTimer;


type
  TStatKind = (sk_Tmp, sk_SortSens, sk_SortInsens, sk_EquiDistSens, sk_EquiDistInsens);

  TFuncType = (ft_ShortSens, ft_ShortInsens, ft_SysSens, ft_SysInsens,
               ft_StatSens, ft_StatInsens, ft_Empty);

  TCompStat = record
    Matches,                                  // counting Mismatches by break positions
    MisMatches    : Array[1..255] of Int64;   // counting Matches by lengths
    NumMatches,
    NumMisMatches,
    NumCompares,
    NumErrors        : Int64;
    TotalTime        : Extended;
  end;

  TStatistics = Array[TStatKind] of TCompStat;

  TSysCompFunc   = function(const S1, S2: AnsiString): LongInt;
  TStatCompFunc  = function(const S1, S2: ShortString; var Stat : TCompStat): SmallInt;

  { TTestSortHashList }

  TTestSortHashList = class(TSortableHashList)
  private
    FSysCompFunc   : TSysCompFunc;
    FStatCompFunc  : TStatCompFunc;
    FStatistics    : TStatistics;
    procedure ResetStatistics(SK : TStatKind);
    procedure QuickSortSysComp(L, R: integer);
    procedure QuickSortStatistics(L, R: integer; SK: TStatKind);
    procedure QuickSortDebug(L, R: integer);
    function  DebugCompare(const S1, S2: ShortString): SmallInt;
    procedure EquiDistCompare(FT : TFuncType; Loops : Int64);
  public
    procedure SortSysCompare(CS : TSHLSensitivity);
    procedure SortStatistics;
    procedure SortDebug(CS : TSHLSensitivity);
    procedure TestPerformance(FT : TFuncType; Duration : integer);
    property  Statistics: TStatistics read FStatistics;
  end;



implementation


{ *************************** Empty Compare function ************************** }

function EmptyCompFunction(const s1, s2: ShortString): SmallInt;
begin
    { Dummy function for measuring the time part needed for parameter loading,
      function call and returning }
end;


{ ******************** Statistics-logging Compare functions ******************* }

function CompStatSensitive(const S1, S2 : ShortString; var Stat : TCompStat) : SmallInt;
var
  P1, P2,
  P1Last : PByte;
  cnt    : LongInt;
begin
  inc(Stat.NumCompares);                    // grabbing statistics
  P1 := @S1;
  P2 := @S2;
  if P1^ > P2^ then P1Last := P1 + P2^ else P1Last := P1 + P1^;
  cnt := 0;                                 // grabbing statistics
  while P1 < P1Last do
  begin
    inc(P1);
    inc(P2);
    inc(cnt);                               // grabbing statistics
    if P1^ <> P2^ then
    begin
      result := P1^ - P2^;
      inc(Stat.Mismatches[cnt]);            // grabbing statistics
      inc(Stat.NumMismatches);              // grabbing statistics
      exit;
    end;
  end;
  result := Length(S1) - Length(S2);
  inc(Stat.Matches[cnt]);                   // grabbing statistics
  inc(Stat.NumMatches);                     // grabbing statistics
end;

function CompStatInsensitive(const S1, S2 : ShortString; var Stat : TCompStat) : SmallInt;
var
  B1, B2         : Byte;
  P1, P2, P1Last : PByte;
  cnt            : LongInt;
begin
  inc(Stat.NumCompares);                    // grabbing statistics
  P1 := @S1;
  P2 := @S2;
  if P1^ > P2^ then P1Last := P1 + P2^ else P1Last := P1 + P1^;
  cnt := 0;                                 // grabbing statistics
  while P1 < P1Last do
  begin
    inc(P1);
    inc(P2);
    inc(cnt);                               // grabbing statistics
    if P1^ <> P2^ then
    begin
      B1 := P1^;
      B2 := P2^;
      if B1 in [97..122] then dec(B1, 32);
      if B2 in [97..122] then dec(B2, 32);
      if B1 <> B2 then
      begin
        result := B1 - B2;
        inc(Stat.Mismatches[cnt]);          // grabbing statistics
        inc(Stat.NumMismatches);            // grabbing statistics
        exit;
      end;
    end;
  end;
  result := Length(S1) - Length(S2);
  inc(Stat.Matches[cnt]);                   // grabbing statistics
  inc(Stat.NumMatches);                     // grabbing statistics
end;


{ TTestSortHashList }


{ ********************** Use of SysUtils CompareStr/Text ********************** }

procedure TTestSortHashList.QuickSortSysComp(L, R : integer);
{ Just a copy of TSortableHashList.QuickSort witch calles TAnsiCompFunc instead
  of TShortCompFunc }
var
  I, J, P        : Integer;
  tmp, PivotItem : PShortString;
begin
  repeat
    I := L;
    J := R;
    P := (L + R) div 2;
    repeat
      PivotItem := SortArray[P];
      while FSysCompFunc(PivotItem^, SortArray[I]^) > 0 do Inc(I);
      while FSysCompFunc(PivotItem^, SortArray[J]^) < 0 do Dec(J);
      if I <= J then
      begin
        tmp := SortArray[i];
        SortArray[I] := SortArray[J];
        SortArray[J] := tmp;
        if P = I then P := J
        else if P = J then P := I;
        Inc(I);
        Dec(J);
      end;
    until I > J;
    if L < J then QuickSortSysComp(L, J);
    L := I;
  until I >= R;
end;

procedure TTestSortHashList.SortSysCompare(CS : TSHLSensitivity);
begin
  ResetSortArray;
  if (Count < 2) then exit;
  CaseSensitivity := CS;
  if CS = CaseSensitive then FSysCompFunc := @CompareStr
  else FSysCompFunc := @CompareText;
  QuickSortSysComp(0, Count-1);
end;


{ *************** Grabbing Statistics (Sort & Equidistribution) *************** }

procedure TTestSortHashList.ResetStatistics(SK : TStatKind);
begin
  with Statistics[SK] do
  begin
    FillChar(Matches[1], Length(Matches) * SizeOf(Int64), #0);
    FillChar(Mismatches[1], Length(Mismatches) * SizeOf(Int64), #0);
    NumMismatches := 0;
    NumMatches    := 0;
    NumCompares   := 0;
    NumErrors     := 0;
    TotalTime     := 0;
  end;
end;

procedure TTestSortHashList.QuickSortStatistics(L, R : integer; SK : TStatKind);
{ Just a copy of TSortableHashList.QuickSort witch deals with TStatCompFunc instead
  of TShortCompFunc }
var
  I, J, P        : Integer;
  tmp, PivotItem : PShortString;
begin
  repeat
    I := L;
    J := R;
    P := (L + R) div 2;
    repeat
      PivotItem := SortArray[P];
      while FStatCompFunc(PivotItem^, SortArray[I]^, FStatistics[SK]) > 0 do Inc(I);
      while FStatCompFunc(PivotItem^, SortArray[J]^, FStatistics[SK]) < 0 do Dec(J);
      if I <= J then
      begin
        tmp := SortArray[i];
        SortArray[I] := SortArray[J];
        SortArray[J] := tmp;
        if P = I then P := J
        else if P = J then P := I;
        Inc(I);
        Dec(J);
      end;
    until I > J;
    if L < J then QuickSortStatistics(L, J, SK);
    L := I;
  until I >= R;
end;

procedure TTestSortHashList.SortStatistics;
begin
  ResetSortArray;
  if (Count < 2) then exit;
  FStatCompFunc := @CompStatSensitive;
  QuickSortStatistics(0, Count-1, sk_SortSens);
  ResetSortArray;
  FStatCompFunc := @CompStatInsensitive;
  QuickSortStatistics(0, Count-1, sk_SortInsens);
end;


{ ***************************** Compare Debugging ***************************** }

function TTestSortHashList.DebugCompare(const S1, S2 : ShortString) : SmallInt;
var
  SysCompResult,
  ShortCompResult,
  StatCompResult     : SmallInt;
  cnt                : LongInt;

  {Sub-}
  function NormalizeResult(CompRes: integer): integer;
  begin
    if CompRes > 0 then Result := 1         // Compare result values besides =0, <0, >0 are
    else if CompRes < 0 then Result := -1   // undefined, so results may differ in value,
         else Result := 0;                  // but must match 0 xor sign
  end;
  {/Sub}

begin
  if CaseSensitivity = CaseSensitive then
  begin
    ShortCompResult := NormalizeResult(CompareShortStringSensitive(s1, s2));
    SysCompResult   := NormalizeResult(CompareStr(s1, s2));
    StatCompResult  := NormalizeResult(CompStatSensitive(s1, s2, FStatistics[sk_Tmp]));
  end
  else begin
    ShortCompResult := NormalizeResult(CompareShortStringInsensitive(s1, s2));
    SysCompResult   := NormalizeResult(CompareText(s1, s2));
    StatCompResult  := NormalizeResult(CompStatInsensitive(s1, s2, FStatistics[sk_Tmp]));
  end;
  if (SysCompResult <> ShortCompResult) or (SysCompResult <> StatCompResult) then
  begin
    inc(FStatistics[sk_Tmp].NumErrors);
    for cnt := 1 to 10 do
      begin
        {*******************************************************
        *  Breakpoint position: 10x loop on deviating results  *
        *******************************************************}
        ShortCompResult := CompareShortStringSensitive(s1, s2);
        StatCompResult  := CompStatSensitive(s1, s2, FStatistics[sk_Tmp]);
        ShortCompResult := CompareShortStringInsensitive(s1, s2);
        StatCompResult  := CompStatInsensitive(s1, s2, FStatistics[sk_Tmp]);
      end;
  end;
  result := SysCompResult;
end;

procedure TTestSortHashList.QuickSortDebug(L, R : integer);
{ Just a copy of TSortableHashList.QuickSort witch deals with DebugCompare instead
  of TShortCompFunc }
var
  I, J, P        : Integer;
  tmp, PivotItem : PShortString;
begin
  repeat
    I := L;
    J := R;
    P := (L + R) div 2;
    repeat
      PivotItem := SortArray[P];
      while DebugCompare(PivotItem^, SortArray[I]^) > 0 do Inc(I);
      while DebugCompare(PivotItem^, SortArray[J]^) < 0 do Dec(J);
      if I <= J then
      begin
        tmp := SortArray[i];
        SortArray[I] := SortArray[J];
        SortArray[J] := tmp;
        if P = I then P := J
        else if P = J then P := I;
        Inc(I);
        Dec(J);
      end;
    until I > J;
    if L < J then QuickSortDebug(L, J);
    L := I;
  until I >= R;
end;

procedure TTestSortHashList.SortDebug(CS : TSHLSensitivity);
begin
  ResetSortArray;
  ResetStatistics(sk_Tmp);
  if (Count < 2) then exit;
  CaseSensitivity := CS;
  QuickSortDebug(0, Count-1);
end;


{ ********************* Performance Test Equidistribution ********************* }

procedure TTestSortHashList.EquiDistCompare(FT : TFuncType; Loops : Int64);
var
  j, k, r : LongInt;
begin
  j := 0;
  k := Count - 1;
  ;
  FStatistics[sk_Tmp].NumCompares := Loops;
  if FT = ft_StatSens then ResetStatistics(sk_EquiDistSens);
  if FT = ft_StatInsens then ResetStatistics(sk_EquiDistInsens);
  DebugTimerStart;
  while Loops > 0 do
  begin
    case FT of
      ft_ShortSens   : r := CompareShortStringSensitive(SortArray[j]^, SortArray[k]^);
      ft_ShortInsens : r := CompareShortStringInsensitive(SortArray[j]^, SortArray[k]^);
      ft_SysSens     : r := CompareStr(SortArray[j]^, SortArray[k]^);
      ft_SysInsens   : r := CompareText(SortArray[j]^, SortArray[k]^);
      ft_StatSens    : r := CompStatSensitive(SortArray[j]^, SortArray[k]^,
                                                     FStatistics[sk_EquiDistSens]);
      ft_StatInsens  : r := CompStatInsensitive(SortArray[j]^, SortArray[k]^,
                                                     FStatistics[sk_EquiDistInsens]);
      ft_Empty       : r := EmptyCompFunction(SortArray[j]^, SortArray[k]^);
    end;
    dec(Loops);
    if j < (Count - 7) then inc(j) else j := 0;
    if k > 0 then dec(k) else k := Count - 1;
  end;
  FStatistics[sk_Tmp].TotalTime := DebugTimerStop;
end;

procedure TTestSortHashList.TestPerformance(FT : TFuncType; Duration : integer);
const
  TenE8 = 100000000;
var
  Loops : Int64;
begin
  ResetSortArray;
  EquiDistCompare(FT, TenE8);
  Loops := round(Duration * 10 / FStatistics[sk_Tmp].TotalTime) * TenE8 div 10;
  EquiDistCompare(FT, Loops);
end;

end.




(*{ ************************** Rejected alternatives ************************** }



{ Just for studying and/or tinkering: Prior development versions, slower than those
  used in unit SortableHashList. }



{ Case-Sensitive - Simple Version: Time consumption 138 % compared to complex pascal version }

function CompareShortStringSensitive(const S1, S2 : ShortString): SmallInt;
var
  i, len : SmallInt;
begin
  if Length(S1) > Length(S2) then len := Length(S2) else len := Length(S1);
  i := 1;
  while i <= len do
  begin
    if (S1[i] = S2[i]) then inc(i)
    else begin
      result := Byte(S1[i]) - Byte(S2[i]);
      exit;
    end;
  end;
  result := Length(S1) - Length(S2);
end;



{ Case-Sensitive - Pointer Version: Time consumption 131 % compared to complex pascal version }

function CompareShortStringSensitive(const S1, S2 : ShortString) : SmallInt;
var
  P1, P2, P1Last : PByte;
begin
  if S1[0] > S2[0] then P1Last := P1 + Byte(S2[0]) else P1Last := P1 + Byte(S1[0]);
  P1 := @S1+1;
  P2 := @S2+1;
  while P1 <= P1Last do
  begin
    if P1^ <> P2^ then
    begin
      result := P1^ - P2^;
      exit;
    end;
    inc(P1);
    inc(P2);
  end;
  result := Length(S1) - Length(S2);
end;



{ Case-Insensitive - Simple Version: Time consumption 138 % compared to complex pascal version }

function CompareShortStringInsensitive(const S1, S2 : ShortString) : SmallInt;
var                             // Number and size of local vars affects speed, but counter
  i : SmallInt;                 // var i must be SmallInt instead of Byte, otherwise match with
  B1, B2 : Byte;                // length=255 will cause an endless-loop. Thus "result" is
                                // "abused" as length variable before it's real assignment.
begin
  if Length(S1) > Length(S2) then result := Length(S2) else result := Length(S1);
  i := 1;                          // Length(S1) is coded similar to Byte(S1[0])
  while i <= result do
  begin
    B1 := Byte(S1[i]);             // Copies needed for modification
    B2 := Byte(S2[i]);             // Doing it here instead of inside the else-part accelerates
    if B1 = B2 then inc(i)         // first-char-mismatch and different case matches at the
    else begin                     // expense of long matching runs.
      if B1 in [97..122] then dec(B1, 32);
      if B2 in [97..122] then dec(B2, 32);
      if B1 = B2 then inc(i)       // Sensitivity match, not done
      else begin
        result := B1 - B2;         // Mismatch found, "real" result assignment, now it's done
        exit;                      // Label is faster than Break
      end;
    end;
  end;                                 // Only reached if all chars matched and i>len: In this
  result := Length(S1) - Length(S2);   // case an additional length comparison is needed.
end;



{ Pointer version insensitive, 20 % slower than complex version }

function CompareShortStringInsensitive(const S1, S2 : ShortString) : SmallInt;
var
  B1, B2 : Byte;
  P1, P2, P1Last : PByte;
label L2;

begin
  P1 := @S1;
  P2 := @S2;
  if P1^ > P2^ then P1Last := P1 + P2^ else P1Last := P1 + P1^;
  while P1 < P1Last do
  begin
    inc(P1);
    inc(P2);
    if P1^ <> P2^ then
    begin
      B1 := P1^;
      B2 := P2^;
      if B1 in [97..122] then dec(B1, 32);
      if B2 in [97..122] then dec(B2, 32);
      if B1 <> B2 then
      begin
        result := B1 - B2;
        goto L2;
      end;
    end;
  end;
  result := Length(S1) - Length(S2);
  L2:
end;



{ ******************** Rejected alternative compare_x86_64 ******************* }

{ For studying purposes: This is a (really elegant, but unfortunately 20 % slower
  version of the case-sensitive compare function, using the processors built-in
  string manipulations abilities. It's core consists of a REPE/REPZ CMPSB CPU
  loop instruction at Label .L1, which can be read as "Repeat a Bytewise String
  Compare while Bytes are equal and counter isn't zero".
  On a system using Little Endian order there is no way to use this instruction
  family with wider access than bytewise, as we not only need the equal/unequal
  information, but also a </> decision on mismatch. Ordered to Little Endian the
  most significant mismatched Byte is located at the least significant position,
  so a REPZ CMPSQ/D/W would bring mostly erroneous results.
  It turned out, that this aproach is significantly slower than the more
  "conventional" way using MOVs, Endian swapping (BSWAP) and comparing quadword-,
  dword- and wordwise. Sorting 1.000.000 strings in a TFPHashList using REPE/REPZ CMPSB
  is about 20 % slower than using the more complex function, but keep in mind that
  these timings include also the (constant) part of Quicksort runtime (outside the
  Compare routines), so the real difference between those functions is even
  (presumably much) greater. }

function CompareShortStringSensitive(const S1, S2 : ShortString): SmallInt;
begin
  asm
           mov    S1, %rsi
           mov    S2, %rdi
           cld
           xor    %rcx, %rcx
           xor    %rax, %rax
           movb  (%rdi), %cl
           inc    %rdi
           lodsb
           mov    %cl, %ah
           cmp    %cl, %al
           ja     .L1
           movb   %al, %cl
    .L1:   repz   cmpsb
           je     .L2
           ja     .L3
           mov    $-1, result
           jmp    .LExit
    .L2:   sbb    %ah, %al
           shl    $8, %ax
           mov    %ax, result
           jmp    .LExit
    .L3:   mov    $1, result
    .LExit:
  end ['rsi', 'rdi', 'rax', 'rcx'];
end;



{ ************* Alternative Sort function outside of TFPHashList ************** }


procedure QuickSortHashList(const HL: TFPHashList; var SA : TSortArray;
                                                   CompareFunction: TShortStrCompareFunc);
var
  n : LongInt;

  {Sub-}
  procedure DoQuickSort(L, R: LongInt);    // modified from Free Pascal TFPGHashTable.Sort
  var
    I, J, P        : LongInt;
    tmp, PivotItem : PShortString;
  begin
    repeat
      I := L;
      J := R;
      P := (L + R) div 2;
      repeat
        PivotItem := SA[P];
        while CompareFunction(PivotItem^, SA[I]^) > 0 do Inc(I);
        while CompareFunction(PivotItem^, SA[J]^) < 0 do Dec(J);
        if I <= J then
        begin
          tmp := SA[i];
          SA[I] := SA[J];
          SA[J] := tmp;
          if P = I then P := J
          else if P = J then P := I;
          Inc(I);
          Dec(J);
        end;
      until I > J;
      if L < J then DoQuickSort(L, J);
      L := I;
    until I >= R;
  end;
  {/Subprocedure DoQuickSort}

{main QuickSortHashList}
begin
  if not Assigned(HL) or not Assigned (CompareFunction) then exit;
  SetLength(SA, HL.Count);
  for n := 0 to High(SA) do
    SA[n] := PShortString(@HL.Strs[HL.List^[n].StrIndex]);
  if High(SA) < 2 then exit;                  // due to the special case of Count=1
  DoQuickSort(0, High(SA));
end;

*)
