unit SortableHashList;

(*                                                                             *
  Name:      SortableHashList                                                  *
                                                                               *
  Objective: * Fast HashList ShortString-Key sorting functionality with a      *
               descendant of Free Pascals ContNrs.TFPHashList.                 *
             * Fast case-sensitive and -insensitive ShortString Compare        *
               functions similar to SysUtils.CompareStr and CompareText        *
               in order to accelerate ShortString comparison and sorting.      *
                                                                               *
  Author:    Ruediger Walter <rue.walter@web.de>                               *
                                                                               *
  Version:   0.4.2 / 2014-11-07                                                *
                                                                               *
  Features:  TSortableHashList is a descendant of TFPHashList with the         *
             ability of sorting its keys and to hand them over to StringLists. *
             By nature of the hash algorithms a hashlist cannot be sorted      *
             internally and does not benefit from a sorted order in terms      *
             of searchtime. So the aim of sorting hashlist keys is basically   *
             the hand-over to a user interface, e.g. a ListBox. Typically      *
             this will be done via a Stringlist.                               *
             The sorting of TSortableHashList does not change anything inside  *
             the hashlist - the data themselves are not touched. Instead the   *
             basic idea of TSortableFPHashList is to provide an additional     *
             array to the hashlist, which contains pointers to the hashlists   *
             keys in sorted order.                                             *
                                                                               *
             TFPHashList is designed for speed, thus it's keys are limited     *
             to ShortStrings (if you have to deal with AnsiString keys         *
             instead, you should use one of the several TFPCustomHashTable-    *
             descendants).                                                     *
             The Free pascal RTL does not offer specialized ShortString        *
             compare functions, it's "standard" functions SysUtils.CompareStr  *
             (case-sensitive) and CompareText (case-insensitive) are designed  *
             to deal with AnsiStrings. Any sorting procedure relies heavily    *
             on string comparison functions and will be significantly slowed   *
             down, if the compare function in use needs string type            *
             conversion on every single function call.                         *
             The CompareShortString(In)Sensitive functions are designed to     *
             avoid this string conversion and to deal efficiently with the     *
             internal structure of ShortStrings. Using them instead of the     *
             SysUtils standard routines speeds the hashList sorting up by      *
             a factor of about 3 to 6 (depending on case-sensitivity, data     *
             distribution and CPU).                                            *
                                                                               *
  Usage:     Constructor Create / Property CaseSensitivity: Sorting can be     *
             done either case-sensitive or case-insensitive. Parameters for    *
             creation of TSortableHashList and the property CaseSensitivity    *
             are either "CaseSensitive" or "CaseInsensitive".                  *
                                                                               *
             Procedure Sort / Property SortArray: The Sort procedure does not  *
             affect the internal data structure of the hashlist. First it      *
             writes the PShortString-Pointers of the hashlist keys into a      *
             dynamic PShortString-Array named "SortArray" in the indexed       *
             order of the hashlist. When Sort is done, those Shortstring-      *
             Pointers are redistributed in a sorted order.                     *
             Please note: There is no automatic update of SortArray, so if     *
             you change the hashlists content, SortArray will be out of date   *
             and you have to sort the list again. You should never write       *
             anything to SortArray, there's also no need to initiate or reset  *
             SortArray. Please keep in mind the specific rules of dealing      *
             with hashlist keys: E.g. an instruction like                      *
             >   SortArray[i]^ := 'I dont understand HashLists';               *
             will corrupt the hashlist data structure (in this case you have   *
             to use (TFPHashList.)Rename and call the Sort procedure again).   *
                                                                               *
             Procedure SortToStrings(ST:TStrings; LoadOption:TSHLLoadOption):  *
             Clears the stringlist ST, calls Sort and copies the hashlist      *
             keys in sorted order to the stringlist. There is no need to set   *
             ST.Sorted = true, this would only slow down the loading process.  *
             LoadOption can take the values "NoObjects" or "WithObjects".      *
             With "NoObjects" only the keys are copied to the stringlist,      *
             otherwise pointers to the hashlist items will be written to       *
             ST.Objects additionally. Please note that TStrings expects        *
             TObject-Pointers, while the hashlist accepts any kind of          *
             pointers. Adding those pointers to the stringlist is done via     *
             typecast and without any type check - you should know what you're *
             doing in the case of non-TObject-pointers (e.g. avoid setting     *
             ST.OwnObjects to true in this case).                              *
                                                                               *
             Function CompareShortString(In)Sensitive: Those functions are     *
             ShortString counterparts to the SysUtils-functions CompareStr     *
             and CompareText. They are not convenient for use with "normal"    *
             strings (which are mostly interpreted as AnsiStrings by FPC) -    *
             they explicitely use the internal representation of ShortStrings  *
             and will not work with AnsiStrings. Similar to their SysUtils     *
             counterparts they work on a byte-by-byte comparison basis and     *
             do not support WideChars, UTF-8 etc.                              *
             Compare Results are: S1 < S2 : Result < 0                         *
                                  S1 = S2 : Result = 0                         *
                                  S1 > S2 : Result > 0                         *
             S1 < S2 means:  (S1[n] < S2[n] on mismatch or                     *
                             Length(S1) < Length(S2) if all chars matched.     *
             Please note that the precise values besides equal, < or > are     *
             undefined.                                                        *
                                                                               *
  Limits:    * ShortStrings only                                               *
             * No automatic Sort update                                        *
                                                                               *
  Credits:   TSortableHashList.QuickSort was adapted from TFPGHashTable.Sort,  *
             and the Compare routines have been inspired by analyzing their    *
             SysUtils counterparts. So the real merits are on the side of the  *
             Free Pascal RTL authors!                                          *
                                                                               *
  Copyright (C) 2014 by Ruediger Walter <rue.walter@web.de>                    *
                                                                               *
  This library is licensed on the same Modifyed LGPL as Free Pascal RTL and    *
  Lazarus LCL. See the file COPYING.FPC, included in your FPC/Lazarus          *
  distribution, for details about the copyright.                               *
                                                                               *
  This program is distributed in the hope that it will be useful, but WITHOUT  *
  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or        *
  FITNESS FOR A PARTICULAR PURPOSE.                                            *
                                                                               *)



{$mode objfpc} {$H-}

interface

uses
  Classes, SysUtils, Contnrs;


type
  TShortStrCompareFunc = function(const S1, S2: ShortString): SmallInt;

  TSHLSortArray   = array of PShortString;
  TSHLSensitivity = (CaseSensitive, CaseInsensitive);
  TSHLLoadOption  = (NoObjects, WithObjects);


{ TSortableHashList }

TSortableHashList = class(TFPHashList)
private
  FSortArray   : TSHLSortArray;
  FSensitivity : TSHLSensitivity;
  FCompFunc    : TShortStrCompareFunc;
  procedure   SetSensitivity(Sensitivity : TSHLSensitivity);
  procedure   QuickSort(L, R: integer);
protected
  procedure   ResetSortArray;
public
  constructor Create(Sensitivity : TSHLSensitivity);
  destructor  Destroy; override;
  procedure   Sort;
  procedure   SortToStrings(ST : TStrings; LoadOption : TSHLLoadOption);
  property    SortArray  : TSHLSortArray read FSortArray;
  property    CaseSensitivity : TSHLSensitivity read FSensitivity write SetSensitivity;
end;


{ Case-Sensitive bytewise comparison of ShortStrings }
function CompareShortStringSensitive(const S1, S2 : ShortString): SmallInt;
                       {$ifndef CPUX86_64} {$ifndef CPUI386} inline; {$endif} {$endif}

{ Case-Insensitive bytewise comparison of ShortStrings }
function CompareShortStringInsensitive(const S1, S2 : ShortString) : SmallInt;
                       {$ifndef CPUX86_64} {$ifndef CPUI386} inline; {$endif} {$endif}


implementation


{ ***************************** TSortableHashList ***************************** }

constructor TSortableHashList.Create(Sensitivity : TSHLSensitivity);
begin
  inherited Create;
  SetSensitivity(Sensitivity);
end;

destructor TSortableHashList.Destroy;
begin
  SetLength(FSortArray, 0);
  inherited Destroy;
end;

procedure TSortableHashList.SetSensitivity(Sensitivity : TSHLSensitivity);
begin
  { The usual "if Sensitivity = FCaseOption then exit;" will trigger a SIGSEGV error
    by skipping the FCompFunc-assignemnt from Create, if Sensitivity = CaseSensitive
    (which internally is handled as zero, to which FCaseOption is initiated on creation)!
    Otherwise the following lines must be doubled in TSortableFPHashList.Create. }
  FSensitivity := Sensitivity;
  if Sensitivity = CaseSensitive then
    FCompFunc := @CompareShortStringSensitive
  else FCompFunc := @CompareShortStringInsensitive;
end;

procedure TSortableHashList.QuickSort(L, R : integer);
var
  I, J, P        : Integer;
  tmp, PivotItem : PShortString;
begin
  repeat
    I := L;
    J := R;
    P := (L + R) div 2;
    repeat
      PivotItem := FSortArray[P];
      while FCompFunc(PivotItem^, FSortArray[I]^) > 0 do Inc(I);
      while FCompFunc(PivotItem^, FSortArray[J]^) < 0 do Dec(J);
      if I <= J then
      begin
        tmp := FSortArray[i];
        FSortArray[I] := FSortArray[J];
        FSortArray[J] := tmp;
        if P = I then P := J
        else if P = J then P := I;
        Inc(I);
        Dec(J);
      end;
    until I > J;
    if L < J then QuickSort(L, J);
    L := I;
  until I >= R;
end;

procedure TSortableHashList.ResetSortArray;
var
  n : integer;
begin
  SetLength(FSortArray, Count);
  for n := 0 to Count - 1 do
    FSortArray[n] := PShortString(@Strs[List^[n].StrIndex]);
end;

procedure TSortableHashList.Sort;
begin
  ResetSortArray;
  if (Count < 2) then exit;
  QuickSort(0, Count-1);
end;

procedure TSortableHashList.SortToStrings(ST : TStrings; LoadOption : TSHLLoadOption);
var
  i : integer;
begin
  if not Assigned(ST) then exit;
  ST.Clear;
  Sort;
  if LoadOption = WithObjects then
    for i := 0 to Count - 1 do
      ST.AddObject(FSortArray[i]^, TObject(Find(FSortArray[i]^)))
  else for i := 0 to Count - 1 do
    ST.Add(FSortArray[i]^);
end;


{ *********************** ShortString Compare functions *********************** }


{$ASMMODE att}
{$ifdef CPUX86_64}                                   { x86-64 Assembler version }

{ For case-sensitive compare full register block comparison can be used. But because
  we need a lower/greater-decision (if not equal; the precise result values besides
  that are undefined) the string block must be loaded into the registers in a Big-Endian
  order. As the x86_64 CPUs use Little-Endian, an Endian swapping must taking place
  before any block comparison (the first string mismatch from left to right must be
  at the most significant register position). For the same reason the specialized CPU
  string instructions (REPE/REPZ CMPSQ/D/W) cannot be used (and using them bytewise
  (CMPSB) is way too slow). }

  function CompareShortStringSensitive(const S1, S2 : ShortString): SmallInt;
  begin
    asm                              { Performance: 26 % faster than best .pas version }
  { Preliminaries: Register preparation }
             movq   S1, %rdi         // @S1 -> RDI
             movq   S2, %rsi         // @S2 -> RSI
             xor    %rcx, %rcx       // Clear RCX, used as counter register
  { Length evaluation }
             movzbw (%rdi), %ax      // Length-Byte(S1) -> AX
             movzbw (%rsi), %dx      // Length-Byte(S2) -> DX
             inc    %rdi             // Increment @S1 / @S2 in RDI / RSI to the first Char
             inc    %rsi
             movw   %dx, %cx         // Length(S2) -> CX
             cmp    %dx, %ax         // Compare (Length(S1) in AX - Length(S2) in DX)
             jae    .L1              // if Length(S2) <= Length(S1) then CX is okay
             movw   %ax, %cx         //   else Length(S1) -> CX;
      .L1:   sub    %dx, %ax         // (Length(S1) - Length(S2)) -> AX
             movw   %ax, Result      // (Length(S1) - Length(S2)) in AX -> Result (needed on match)
  { Now begin with 8 Byte substring Char comparison }
      .L2:   cmp    $8, %cx          // Length >= 8 ?
             jb     .L3              // If Length < 8 then goto L3 (4 Byte comparison)
             movq   (%rdi), %rax     // 8 Byte String S1 from RDI -> RAX (faster than LODSQ)
             movq   (%rsi), %rdx     // 8 Byte String S2 from RSI -> RDX
             bswap  %rax             // RAX: Little Endian -> Big Endian
             bswap  %rdx             // RDX: Little Endian -> Big Endian
             cmp    %rdx, %rax       // Now compare whole 8 Byte array (RAX - RDX) = (S1 - S2)
             ja     .L7              // if Mismatch and S1 > S2 then goto L7 where Result := +1
             jb     .L6              // if Mismatch and S1 < S2 then goto L6 where Result := -1
             add    $8, %rdi         // inc(String Pointer RDI & RSI, 8)
             add    $8, %rsi
             sub    $8, %cx          // dec(Length in CX, 8)
             jmp    .L2              // try again 8 Byte Compare
  { 4 Byte substring comparison, there can't be more than 7 Chars left }
      .L3:   cmp    $4, %cx          // Length >= 4 ?
             jb     .L4              // If Length < 4 then goto L4 (2 Byte comparison)
             movl   (%rdi), %eax     // 4 Byte String S1 from RDI -> EAX (faster than LODSD/L)
             movl   (%rsi), %edx     // 4 Byte String S2 from RSI -> EDX
             bswap  %eax             // EAX: Little Endian -> Big Endian
             bswap  %edx             // EDX: Little Endian -> Big Endian
             cmp    %edx, %eax       // Compare 4 Byte EAX - EDX = S1 - S2
             ja     .L7              // if Mismatch and S1 > S2 then goto L7 where Result := +1
             jb     .L6              // if Mismatch and S1 < S2 then goto L6 where Result := -1
             add    $4, %rdi         // inc(String Pointer RDI & RSI, 4)
             add    $4, %rsi
             sub    $4, %cx          // dec(Length in CX, 4)
  { 2 Byte substring comparison, there can't be more than 3 Chars left }
      .L4:   cmp    $2, %cx          // Length >= 2 ?
             jb     .L5              // If Length < 2 then goto L5 (last Byte comparison)
             movw   (%rdi), %ax      // 2 Byte String S1 from RDI -> AX
             movw   (%rsi), %dx      // 2 Byte String S2 from RSI -> DX
             cmp    %dl, %al         // Compare AL - DL Byte first due to Little Endian
             ja     .L7              // if Mismatch and S1 > S2 then goto L7 where Result := +1
             jb     .L6              // if Mismatch and S1 < S2 then goto L6 where Result := -1
             cmp    %dh, %ah         // Compare AH - DH Byte next (= S1 - S2)
             ja     .L7              // if Mismatch and S1 > S2 then goto L7 where Result := +1
             jb     .L6              // if Mismatch and S1 < S2 then goto L6 where Result := -1
             sub    $2, %cx          // dec(Length in CX, 4)
             jz     .Lx              // Exit if no Byte left (all matched, result is already assigned)
             add    $2, %rdi         // inc(String Pointer RDI & RSI, 2)
             add    $2, %rsi
  { Last Byte comparison, there must be 0 or 1 Byte left if reached }
      .L5:   jecxz  .Lx              // No Byte left, all matched, Length result is already assigned
             movb   (%rdi), %al      // Last Byte of S1 from RDI -> AL
             movb   (%rsi), %dl      // Last Byte of S2 from RSI -> DL
             cmp    %dl, %al         // Compare last Byte, no counter update needed any more
             je     .Lx              // Last Byte okay, all matched, Length result is already assigned
             ja     .L7              // if Mismatch and S1 > S2 then goto L7 where Result := +1
  { Mismatch Result assignment }
      .L6:   movw   $-1, result      // if S1 < S2 then Result := -1
             jmp    .Lx              // Done...
      .L7:   movw   $1, result       // if S1 > S2 then Result := 1
      .Lx:                           // Exit
    end ['rax', 'rcx', 'rdx', 'rsi', 'rdi'];
  end;

{ Other than CompareShortStringSensitive the insensitive compare function doesn't
  use Endian swapping. This is because of the necessity to inspect and possibly
  upcase every singe Byte in case of a mismatch. This can't be done blockwise,
  we have to work from right to the left instead. Doing so it's better to have the
  registers in Little Endian order. }

function CompareShortStringInsensitive(const S1, S2 : ShortString) : SmallInt;
Begin
  asm                              { Performance: 24 % faster than best .pas version }
{ Preliminaries: Register preparation }
           movq   S1, %rdi         // @S1 -> RDI
           movq   S2, %rsi         // @S2 -> RSI
           xor    %rcx, %rcx       // Clear RCX, used as counter register
           { Length evaluation }
           movzbw (%rdi), %ax      // Length-Byte(S1) -> AX
           movzbw (%rsi), %dx      // Length-Byte(S2) -> DX
           inc    %rsi             // Increment @S1 in RSI to the first Char
           inc    %rdi             // Increment @S2 in RDI to the first Char
           movw   %dx, %cx         // Length(S2) -> CX
           cmp    %dx, %ax         // Compare (Length(S1) in AX - Length(S2) in DX)
           jge    .L1              // if Length(S2) <= Length(S1) then CX is okay
           movw   %ax, %cx         //   else Length(S1) -> CX;
    .L1:   sub    %dx, %ax         // (Length(S1) - Length(S2)) -> AX
           movw   %ax, Result      // (Length(S1) - Length(S2)) in AX -> Result (needed on match)
{ Now begin with 8 Byte substring comparison }
    .L2:   cmp    $8, %cx          // Length >= 8 ?
           jb     .L3              // If Length < 8 then goto L3 (4 Byte comparison)
           movq   (%rdi), %rax     // 8 Byte String S1 from RDI -> RAX (faster than LODSQ)
           movq   (%rsi), %rdx     // 8 Byte String S2 from RSI -> RDX
           cmp    %rdx, %rax       // Now compare whole 8 Byte substring (RAX - RDX) = (S1 - S2)
           jne    .L3a             // if Mismatch goto L3 without updating counters: Compare lower DWord
           add    $8, %rdi         // QWord has matched, now update counters
           add    $8, %rsi         // inc String Pointers RSI & RDI by 8
           sub    $8, %cx          // dec Length in CX by 8
           jmp    .L2              // If matched then try again 8 Byte Compare
{ 4 Byte substring comparison, but there can be more than 7 Chars left! }
    .L3:   cmp    $4, %cx          // Length >= 4 ?
           jb     .L4              // If Length < 4 then goto L4 (2 Byte comparison)
           movl   (%rdi), %eax     // 4 Byte String S1 from RDI -> EAX (faster than LODSD/L)
           movl   (%rsi), %edx     // 4 Byte String S2 from RSI -> EDX
           // Next 4 Bytes surely not out of String and loaded in EAX/EDX, but counters not yet updated
    .L3a:  cmp    %edx, %eax       // Compare 4 Byte EAX - EDX = S1 - S2
           jne    .L4a             // if Mismatch goto L4 without updating counters: Word comparison
           add    $4, %rdi         // DWord has matched, now update counters
           add    $4, %rsi         // inc String Pointers RSI & RDI by 4
           sub    $4, %cx          // dec Length in CX by 4
           // The reason why it reached this point is either a mismatch step-down or because it's
           // near the end of the string. So jumping back to big array compare is wasting time!
           jz     .Lx              // But perhaps it's all done ?
{ 2 Byte substring comparison, but there might be more than 3 Chars left! }
    .L4:   cmp    $2, %cx          // Length >= 2 ?
           jb     .L5              // If Length < 2 then goto L5 (last Byte comparison)
           movw   (%rdi), %ax      // 2 Byte String S1 from RDI -> AX and inc(RSI, 4)
           movw   (%rsi), %dx      // 2 Byte String S2 from RSI -> DX
           // Next 2 Bytes surely not out of String and loaded in AX/DX, but counters not yet updated
    .L4a:  cmp    %dx, %ax         // Compare 2 Byte AX - DX = S1 - S2
           jne    .L4b             // on Mismatch goto L4b without updating counters: Which Byte ?
           add    $2, %rdi         // Word has matched, now update counters
           add    $2, %rsi         // inc String Pointers RSI & RDI by 2
           subw   $2, %cx          // dec Length in CX by 2: Are there any Bytes left?
           jz     .Lx              // If not, it's done, Length comparison result is already assigned
           // The reason why it reached this point is either a mismatch step-down or because it's
           // near the end of the string. In the first case the mismatch must be in the next 2 Bytes
           jmp    .L4              // therefore if there are Bytes left, try again 2 Byte Compare
           // Here a mismatch is detected in AX/DX, but where is it?
    .L4b:  cmp    %dl, %al         // First compare the lower Byte
           jne    .Lc              // If mismatch in AL/DL goto .Lc, registers & flags are already set
           inc    %rdi             // Now the mismatch must be in AH/DH,
           inc    %rsi             //   so update counters,
           dec    %cx
           shr    $8, %ax          //   shift H -> L
           shr    $8, %dx
           jmp    .Lc              //   and goto .Lc for case match comparison
{ Single Byte comparison, here can't be anything else than no or the (unloaded) last Char left! }
    .L5:   jecxz  .Lx              // No Byte left, all matched, Length result is already assigned
           movb   (%rdi), %al      // Next Byte from String S1 in RDI -> AL and inc(RSI, 4)
           movb   (%rsi), %dl      // Next Byte from String S2 in RSI -> DL
           cmpb   %dl, %al         // Compare last Byte
           je     .Lx              // if matched, it's done, Length comparison result is already set
{ Case comparison: Here's a mismatch in AL/DL with correct counter positions set, but contrary
  to the position at Label .L5 we don't know whether it's the last Byte or not! So if there is a
  case match, we have to jump back to .L2 (quad-word comparison). In fact, this is the only way
  a different-case match tracks back to block comparison. }
    .Lc:   cmpb   $97, %al         // Eliminating case differences by upcasing chars in AL/DL:
           jb     .Lc1             // if Char in AL >= "a"...
           cmpb   $122, %al
           ja     .Lc1             // and Char in AL <= "z" ...
           subb   $32, %al         // then dec(AL, 32)
    .Lc1:  cmpb   $97, %dl
           jb     .Lc2             // Now do the same with DL
           cmpb   $122, %dl
           ja     .Lc2
           subb   $32, %dl
    .Lc2:  cmp    %dl, %al         // Comparing the (possibly) upcased chars:
           jne    .L6              // On mismatch goto result assignment
           inc    %rdi             // On match update counters
           inc    %rsi             //   (In most cases this is not the end of the line!)
           dec    %cx
           jmp    .L2              // and jump back to .L2 quad-word comparison
{ Mismatch Result assignment }
    .L6:   ja     .L7              // Mismatch: if S1 > S2 then goto L7, where Result := +1
           mov    $-1, result      // else Result := -1
           jmp    .Lx              // Done...
    .L7:   mov    $1, result       // Result := +1
    .Lx:                           // Exit
  end ['rax', 'rcx', 'rdx', 'rsi', 'rdi'];
end;

{$else (ifndef CPUX86_64)}
{$ifdef CPUI386}                                      { i386 Assembler version }

function CompareShortStringSensitive(const S1, S2 : ShortString): SmallInt;
begin
  asm
{ Preliminaries: Register preparation }
           movl   S1, %edi         // @S1 -> EDI
           movl   S2, %esi         // @S2 -> ESI
           xor    %ecx, %ecx       // Clear ECX, used as counter register
{ Length evaluation }
           movzbw (%edi), %ax      // Length-Byte(S1) -> AX
           movzbw (%esi), %dx      // Length-Byte(S2) -> DX
           inc    %edi             // Increment @S1 in EDI to the first Char
           inc    %esi             // Increment @S2 in ESI to the first Char
           movw   %dx, %cx         // Length(S2) -> CX
           cmp    %dx, %ax         // Compare (Length(S1) in AX - Length(S2) in DX)
           jae    .L1              // if Length(S2) <= Length(S1) then CX is okay
           movw   %ax, %cx         //   else Length(S1) -> CX;
    .L1:   sub    %dx, %ax         // (Length(S1) - Length(S2)) -> AX
           movw   %ax, Result      // (Length(S1) - Length(S2)) in AX -> Result (needed on match)
{ Begin with 4 Byte substring comparison }
    .L2:   cmp    $4, %cx          // Length >= 4 ?
           jb     .L3              // If Length < 4 then goto L4 (2 Byte comparison)
           movl   (%edi), %eax     // 4 Byte String S1 from EDI -> EAX (faster than LODSD/L)
           movl   (%esi), %edx     // 4 Byte String S2 from ESI -> EDX
           bswap  %eax             // EAX: Little Endian -> Big Endian
           bswap  %edx             // EDX: Little Endian -> Big Endian
           cmp    %edx, %eax       // Compare 4 Byte EAX - EDX = S1 - S2
           ja     .L6              // if Mismatch and S1 > S2 then goto L6 where Result := +1
           jb     .L5              // if Mismatch and S1 < S2 then goto L5 where Result := -1
           add    $4, %edi         // inc(String Pointer EDI, 4)
           add    $4, %esi         //   and inc(String Pointer ESI, 4)
           sub    $4, %cx          // dec(Length in CX, 4)
           jmp    .L2              // try again 8 Byte Compare
{ 2 Byte substring comparison, there can't be more than 3 Chars left }
    .L3:   cmp    $2, %cx          // Length >= 2 ?
           jb     .L4              // If Length < 2 then goto L4 (last Byte comparison)
           movw   (%edi), %ax      // 2 Byte String S1 from EDI -> AX
           movw   (%esi), %dx      // 2 Byte String S2 from ESI -> DX
           cmp    %dl, %al         // Compare AL - DL Byte first due to Little Endian
           ja     .L6              // if Mismatch and S1 > S2 then goto L6 where Result := +1
           jb     .L5              // if Mismatch and S1 < S2 then goto L5 where Result := -1
           cmp    %dh, %ah         // Compare AH - DH Byte next (= S1 - S2)
           ja     .L6              // if Mismatch and S1 > S2 then goto L6 where Result := +1
           jb     .L5              // if Mismatch and S1 < S2 then goto L5 where Result := -1
           sub    $2, %cx          // dec(Length in CX, 4)
           jz     .Lx              // Exit if no Byte left (all matched, result is already assigned)
           add    $2, %edi         // inc(String Pointer EDI, 2)
           add    $2, %esi         //   and inc(String Pointer ESI, 2)
{ Last Byte comparison, there must be 0 or 1 Byte left if reached }
    .L4:   jecxz  .Lx              // No Byte left, all matched, Length result is already assigned
           movb   (%edi), %al      // Last Byte of S1 from EDI -> AL
           movb   (%esi), %dl      // Last Byte of S2 from ESI -> DL
           cmp    %dl, %al         // Compare last Byte, no counter update needed any more
           je     .Lx              // Last Byte okay, all matched, Length result is already assigned
           ja     .L6              // if Mismatch and S1 > S2 then goto L6 where Result := +1
{ Mismatch Result assignment }
    .L5:   movw   $-1, result      // if S1 < S2 then Result := -1
           jmp    .Lx              // Done...
    .L6:   movw   $1, result       // if S1 > S2 then Result := 1
    .Lx:                           // Exit
  end ['eax', 'ecx', 'edx', 'esi', 'edi'];
end;

function CompareShortStringInsensitive(const S1, S2 : ShortString) : SmallInt;
Begin
  asm
{ Preliminaries: Register preparation }
           movl   S1, %edi         // @S1 -> EDI
           movl   S2, %esi         // @S2 -> ESI
           xor    %ecx, %ecx       // Clear ECX, used as counter register
           { Length evaluation }
           movzbw (%edi), %ax      // Length-Byte(S1) -> AX
           movzbw (%esi), %dx      // Length-Byte(S2) -> DX
           inc    %esi             // Increment @S1 in ESI to the first Char
           inc    %edi             // Increment @S2 in EDI to the first Char
           movw   %dx, %cx         // Length(S2) -> CX
           cmp    %dx, %ax         // Compare (Length(S1) in AX - Length(S2) in DX)
           jge    .L1              // if Length(S2) <= Length(S1) then CX is okay
           movw   %ax, %cx         //   else Length(S1) -> CX;
    .L1:   sub    %dx, %ax         // (Length(S1) - Length(S2)) -> AX
           movw   %ax, Result      // (Length(S1) - Length(S2)) in AX -> Result (needed on match)
{ Now begin with 4 Byte substring comparison }
    .L2:   cmp    $4, %cx          // Length >= 4 ?
           jb     .L3              // If Length < 4 then goto L4 (2 Byte comparison)
           movl   (%edi), %eax     // 4 Byte String S1 from EDI -> EAX (faster than LODSD/L)
           movl   (%esi), %edx     // 4 Byte String S2 from ESI -> EDX
           cmp    %edx, %eax       // Now compare whole 4 Byte substring (EAX - EDX) = (S1 - S2)
           jne    .L3a             // if Mismatch goto L3 without updating counters: Compare lower Word
           add    $4, %edi         // DWord has matched, now update counters
           add    $4, %esi         // inc String Pointers ESI & EDI by 4
           sub    $4, %cx          // dec Length in CX by 4
           jmp    .L2              // If matched then try again 4 Byte Compare
{ 2 Byte substring comparison, but there might be more than 3 Chars left! }
    .L3:   cmp    $2, %cx          // Length >= 2 ?
           jb     .L4              // If Length < 2 then goto L4 (last Byte comparison)
           movw   (%edi), %ax      // 2 Byte String S1 from EDI -> AX and inc(RSI, 4)
           movw   (%esi), %dx      // 2 Byte String S2 from ESI -> DX
           // Next 2 Bytes surely not out of String and loaded in AX/DX, but counters not yet updated
    .L3a:  cmp    %dx, %ax         // Compare 2 Byte AX - DX = S1 - S2
           jne    .L3b             // on Mismatch goto L4b without updating counters: Which Byte ?
           add    $2, %edi         // Word has matched, now update counters
           add    $2, %esi         // inc String Pointers ESI & EDI by 2
           subw   $2, %cx          // dec Length in CX by 2: Are there any Bytes left?
           jz     .Lx              // If not, it's done, Length comparison result is already assigned
           // The reason why it reached this point is either a mismatch step-down or because it's
           // near the end of the string. In the first case the mismatch must be in the next 2 Bytes
           jmp    .L3              // therefore if there are Bytes left, try again 2 Byte Compare
           // Here a mismatch is detected in AX/DX, but where is it?
    .L3b:  cmp    %dl, %al         // First compare the lower Byte
           jne    .Lc              // If mismatch in AL/DL goto .Lc, registers & flags are already set
           inc    %edi             // Now the mismatch must be in AH/DH,
           inc    %esi             //   so update counters,
           dec    %cx
           shr    $8, %ax          //   shift H -> L
           shr    $8, %dx
           jmp    .Lc              //   and goto .Lc for case match comparison
{ Single Byte comparison, here can't be anything else than no or the (unloaded) last Char left! }
    .L4:   jecxz  .Lx              // No Byte left, all matched, Length result is already assigned
           movb   (%edi), %al      // Next Byte from String S1 in EDI -> AL and inc(RSI, 4)
           movb   (%esi), %dl      // Next Byte from String S2 in ESI -> DL
           cmpb   %dl, %al         // Compare last Byte
           je     .Lx              // if matched, it's done, Length comparison result is already set
{ Case comparison: Here's a mismatch in AL/DL with correct counter positions set, but contrary
  to the position at Label .L4 we don't know whether it's the last Byte or not! So if there is a
  case match, we have to jump back to .L2 (DWord comparison). In fact, this is the only way
  a different-case match tracks back to block comparison. }
    .Lc:   cmpb   $97, %al         // Eliminating case differences by upcasing chars in AL/DL:
           jb     .Lc1             // if Char in AL < "a" ...
           cmpb   $122, %al        // ... or ...
           ja     .Lc1             // if Char in AL > "z" do nothing
           subb   $32, %al         // else dec(AL, 32)
    .Lc1:  cmpb   $97, %dl
           jb     .Lc2             // Now do the same with DL
           cmpb   $122, %dl
           ja     .Lc2
           subb   $32, %dl
    .Lc2:  cmp    %dl, %al         // Comparing the (possibly) upcased chars:
           jne    .L5              // On mismatch goto result assignment
           inc    %edi             // On match update counters
           inc    %esi             //   (In most cases this is not the end of the line!)
           dec    %cx
           jmp    .L2              // and jump back to .L2 DWord comparison
{ Mismatch Result assignment }
    .L5:   ja     .L6              // Mismatch: if S1 > S2 then goto L7, where Result := +1
           mov    $-1, result      // else Result := -1
           jmp    .Lx              // Done...
    .L6:   mov    $1, result       // Result := +1
    .Lx:                           // Exit
  end ['eax', 'ecx', 'edx', 'esi', 'edi'];
end;

{$else (ifndef (CPUX86_64 or CPUI386))}               { Pascal version }

function CompareShortStringSensitive(const S1, S2 : ShortString): SmallInt;
var
  i, len   : SmallInt;                                // i must be 16 Bit if match has length=255
  dw1, dw2 : DWord;
begin
  if Length(S1) > Length(S2) then len := Length(S2)   // len = Length(shorter String)
  else len := Length(S1);                             // Same code for Length(S1) & Byte(S1[0])
  { PART 1: Bytewise comparison of Chars 1 & 2. This affects about 40-50 % of all comparisons
    (with Quicksort) and speeds up the search by about 7.5 % compared to a DWord-wise comparison
    from the beginning (6 % comes from the first Byte comparison, 1.5 % from the second). Doing
    it more than twice, wouldn't bring additional benefit. }
  if len > 1 then
  begin
    if S1[1] <> S2[1] then begin result := Byte(S1[1]) - Byte(S2[1]); exit; end;
    if S1[2] <> S2[2] then begin result := Byte(S1[2]) - Byte(S2[2]); exit; end;
    i := 3;
  end
  else i := 1;
  { PART 2: DWord-wise comparison. This speed up about 2/3 of the remaining comparisons
    significantly while moderately slowing down the rest. Overall this accelerates the sorting
    by roundabout 30 %. }
  while i < (len - 2) do
  begin
    if PDWord(@S1[i])^ = PDWord(@S2[i])^ then         // if equal, endianess doesn't matter
      inc(i,4)
    else begin                                        // else Mismatch found
      {$IFDEF ENDIAN_LITTLE}
      dw1 := SwapEndian(PDWord(@S1[i])^);             // Big Endian needed for </> decision
      dw2 := SwapEndian(PDWord(@S2[i])^);             // Copies needed for modification (swapping)
      if dw1 < dw2 then result := -1 else result := 1;
      {$ELSE}
      if PDWord(@S1[i])^ < PDWord(@S2[i])^ then result := -1 else result := 1;
      {$ENDIF}
      exit;                                           // It's done, no length comparison needed
    end;
  end;
  { PART 3: The last 1 - 3 Chars must be treated bytewise. Only about 4 % of all comparisons
    will reach this point. }
  while i <= len do
  begin
    if (S1[i] = S2[i]) then inc(i)
    else begin
      result := Byte(S1[i]) - Byte(S2[i]);    // Mismatch found, no length comparison needed
      exit;
    end;
  end;                                        // Only reached if all chars matched and i>len
  result := Length(S1) - Length(S2);          // Then an additional length comparison is needed.
end;

function CompareShortStringInsensitive(const S1, S2 : ShortString) : SmallInt;
var                                           // i must be 16 Bit if match has length=255
  i, len : SmallInt;
  B1, B2 : Byte;
begin
  if Length(S1) > Length(S2) then len := Length(S2) else len := Length(S1);
  { PART 1: Bytewise comparison of Chars 1 - 5. This affects about 50-60 % of all comparisons
    (with Quicksort) and might speeds up the search by about 8 % compared to a DWord-wise comparison
    from the beginning. But this would be really ugly code, repeating the 12 line sequence over
    and over again with hard coded memory access. There is no way to shorten it while keeping
    the speed advantage. So I desist from that and begin with...}
  i := 1;
  { PART 2: DWord-wise comparison. This speed up about 2/3 of the remaining comparisons
    significantly while moderately slowing down the rest. Overall this accelerates the sorting
    by roundabout 20 %. }
  while i < (len - 2) do
  begin
    if PDWord(@S1[i])^ = PDWord(@S2[i])^ then inc(i,4)      // if equal, endianess doesn't matter
    else begin
      while S1[i] = S2[i] do inc(i);         // no range check needed, break condition must come
      B1 := Byte(S1[i]);                     // now S1[i] <> S2[i], but may case match
      B2 := Byte(S2[i]);
      if B1 in [97..122] then dec(B1, 32);
      if B2 in [97..122] then dec(B2, 32);
      if B1 = B2 then inc(i)                 // Case match, not yet done
      else begin
        result := B1 - B2;                   // Mismatch found, result assignment, now it's done
        exit;
      end;
    end;
  end;
{ PART 3: The last 1 - 3 Chars must be treated bytewise. Only about 4 % of all comparisons
  will reach this point. }
  while i <= len do
  begin
    if S1[i] = S2[i] then inc(i)
    else begin
      B1 := Byte(S1[i]);
      B2 := Byte(S2[i]);
      if B1 in [97..122] then dec(B1, 32);
      if B2 in [97..122] then dec(B2, 32);
      if B1 = B2 then inc(i)                 // Sensitivity match, not yet done
      else begin
        result := B1 - B2;                   // Mismatch found, result assignment, now it's done
        exit;
      end;
    end;
  end;                                 // Only reached if all chars matched and i>len: In this
  result := Length(S1) - Length(S2);   // case an additional length comparison is needed.
end;

{$endif CPUI386}
{$endif CPUX86_64}


end.


